# BOSCH IoT INSIGHTS - Processing Pipeline - Custom Code Sample with PYTHON

## Table of Contents
1. [ Structure of the example                 ]( <#structure>      )
2. [ Setup locale development environment     ]( <#local-dev-env>  )
3. [ How to extend the starter example        ]( <#extend-further> )
4. [ Run the code locally                     ]( <#run-locally>    )
5. [ Deploy the code to Insights              ]( <#deploy>         )
6. [ The exchange data format                 ]( <#data-format>    )
7. [ Glossary                                 ]( <#glossary>       )

## Structure of the example                                                                        <a name="structure"></a>
The example contains these files:

1. __executable-manifest.yaml__ `required`  
  Technical description of custom artifact for Insight's processing pipeline executor.  
  In the manifest you need to configure the dependencies you need for your script and which
  script and class should be used for the startup.
  
2. __src/step.py__ `required` `MAINSCRIPT`  
  This is your Python script with the custom processing logic.  
  You can also have more than one script, but the executable-manifest names exactly one as
  entry point.

3. __src/insights_protocol.py__ `required`  
  This is an example implementation to read and write the ExchangeDataHolder JSON structure.
  It is imported into `step.py` to remove all the encoding and decoding functions from 
  the main script.

4. __README.md__  `docu`  
  Documentation of the source project.

5. __dev-requirements.txt__ `config`  
  This file is used to speed up the setup of the locale development environment.
  With this file you can install all required dependencies of our example:  
  `python -m pip install -r dev-requirements.txt`  
  See section _Setup local development environment_ below for more details.

6. __requirements.txt__ `optional`  
  With this file gives control about Python module version your step depends on.
  See section _How to extend the starter example_ below for more details.

7. __tests/SampleJson.json__ `example`  
  An example input file with JSON, like the step gets, if he executes after a parser step.

8. __tests/SampleBinary.json__ `example`  
  An example input file with binary payload, typical if current step is the first in processing.
  
9. __deploy_artifact.py__ `tool`  
  A Python script to help with fast packaging and deployment of your custom processing code.

10. __test.py__ `tool`  
  Python scripts to help with testing in a local environment.
  See section _Run the code locally_ for more details.

All as `required` marked files are needed in the resulting ZIP file, to make the example is working. The other
files in our example are not required and are just information to help you get started.

## Setup locale development environment                                                            <a name="local-dev-env"></a>

To run the example locally you need to install Python 3.9.

For installation of Python see <https://wiki.python.org/moin/BeginnersGuide/Download>.  
If you have installed Python 3.9 there should be also `pip` included, which is used 
to install required dependencies [Python modules from Python Package Index]( <https://pypi.org> ).

In the case you are not familiar with Python you can have a look at <https://wiki.python.org/moin/BeginnersGuide>.

Now you can install the dependencies for the example.  
As first step unzip the example into a folder and change into the directory.

The folder should look like:

```sh
$ ls -1
README.md
deploy_artifact.py
dev-requirements.txt
executable-manifest.yaml
requirements.txt
src
test.py
tests
```

Depending on your environment one of the following commands should install
the required dependencies:

```sh
$ python3 -m pip install -r dev-requirements.txt
$ python -m pip install -r dev-requirements.txt
```

📝python3 vs. python?
> Many Linux distributions are currently in the process of migrating from
> Python 2 to Python 3.
>
> To keep scripts running, that were written for Python 2, without the need
> to patch or change them, many decided to keep the `python` command
> pointing to the Python 2 installation, and added `python3` as the command
> to run the new scripts, which use Python 3.
>
> Insights steps must use Python 3. This Readme uses `python` in all examples
> and explanations; on your machine, you may have to use `python3` (and 
> `pip3`) instead.

If `pip` is not yet available, the Python community offers a shortcut to install
it to your system:
```sh
$ curl https://bootstrap.pypa.io/get-pip.py | python
```

More information about options and ways to install pip:
<https://pip.pypa.io/en/stable/installing/>

If you work on one of the major Linux distributions, descriptions how to install `pip` with their
package management is available:
<https://packaging.python.org/guides/installing-using-linux-tools/#installing-pip-setuptools-wheel-with-linux-package-managers>

📝 Downloading required modules in corporate networks:
> If you are without direct Internet connection because you're behind a coporate proxy,
> you have to configure proxy settings for `pip` before you can install the dependencies:    
> Set environment variable:
>
>     //Linux
>     $ export https_proxy=YourProxyUrl 
>
>     //Windows
>     $ SET HTTPS_PROXY=YourProxyUrl
>
> If your corporate proxy replaces certificates, you can use the `--trusted-host` option
> of `pip` (can be shortened to `--trust`):
>
>     $ curl -k https://bootstrap.pypa.io/get-pip.py | python - --trust files.pythonhosted.org --trust pypi.org
>
>     $ python -m pip install --trust files.pythonhosted.org --trust pypi.org -r requirements.txt

The full documentation for `pip` is available here: https://pip.pypa.io/en/stable/

## How to extend the starter example                                                               <a name="extend-further"></a>
To extend the example you can edit the file `step.py`.
You can even add more Python files to the `src` directory, to have a clean and understandable project
structure.
And you can use modules from the common [Python Package Index]( <https://pypi.org> ).
In Python projects, its common to write the modules they depend on, into a file `requirements.txt`.
On the remote side the dependencies listed there are installed automatically by default.
But you do not need a `requirements.txt`. If your Python step does not contain one, Insights  
step executor will try to autodetect the dependencies during start of your processing 
pipeline.

We recommend to use a `requirements.txt` to specify the module versions, you are using.
If you have created your `requirements.txt` add it to your deployment package via the
`-r` option when using `deploy_artifact.py`:
```sh
$ python deploy_artifact.py -r ./path/to/requirements.txt
```  

### Background info: How steps are embedded into processing
A processing step represents one transformation within a data processing pipeline that 
works on the uploaded data and enriches, transforms, harmonizes and filters the data 
into a target format, typically to be queried and visualized within Iot Insights.

#### Pipeline

The pipeline triggers on arrival of new uploaded documents. The uploaded document is
wrapped in a JSON document (`exchange data format`, see section below) and transferred 
to the first step.

Each step will now transform and change the document and forward it to the next, until 
the final step (typically an output step) is reached:

    (Trigger) ----> (Step 1) ----> (Step 2) ----> (Output) 

To allow processing even huge amount of data, a project can configure a certain pipeline to 
start and run multiple times in parallel.

#### Step
Each single step is started as soon as the pipeline is started. It can do some initialization 
and eventually has to listen on system standard input for documents.

    _ _     _ _       _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _   _ _     _ _     
    \   \   \   \    |s|                                     |s|  \   \   \   \   
     \   \   \   \   |t|    P R O C E S S I N G   S T E P    |t|   \   \   \   \  
      \   \   \   \  |d|                                     |d|    \   \   \   \ 
       )   )   )   ) |i|                                     |o|     )   )   )   )
      /   /   /   /  |n|           +-------------+           |u|    /   /   /   / 
     /   /   /   /   | |           |   stderr    |           |t|   /   /   /   /  
    ´---´   ´---´    +-+-----------+-------------+-----------+-+  ´---´   ´---´   
    
Each pipeline run is then forwarded to this process as an exchange data JSON document in one 
line (end of line is '\n' = byte value 0x0a) and each output line is forwarded to the next step.

Psydo code:
```python
    for line in sys.stdin:
        print(line, flush=True)
```

#### Edge case: Input files to big to load completly into memory
In the case the data is too large to load directly into the memory, you have the possibility to get a s3 link injected.
Be beware that even with this option, small data below 10Kbyte is injected directly.
With this link you can load the data piece by piece, if your use case allows this.
To activate this function, the project must have an active object storage option.
You must also set the 'input' field in the execution manifest to 'insights-payloadRef+json'.
Now the step get the new field 'payloadRef' on the ExchangeDataHolder with a s3 link for big data objects.
As first step you can have a look into the 'read_object_from_s3.py' snippet.

## Run the code locally                                                                            <a name="run-locally"></a>
To check if your code is working you can execute the code locally for the first try:
```sh
$ python test.py default-tests
```   

If all is working fine you should see two JSON documents in the output. They were the 
result of handing over the two sample files in `tests` directory to `step.py` for
processing them.

To specify exactly the files you want the `step.py` to process, there is `run-files`:
```sh
$ python test.py run-files tests/SampleJson.json
```  

This should result in one JSON document printed out, which represents the result of 
processing SampleJson.json.

📝 Note 
> what test.py does is equivalent to
>
> - for a single file (`test.py run-files tests/SampleJson.json`):
>
>         $ cat tests/SampleJson.json | tr -d "\r\n" | python src/step.py | jq
>
> - for multiple files (`test.py run-files tests/*.json`):
>
>         $ ( for test in tests/*.json; do echo $(cat $test | tr -d "\r\n"); done ; ) | python src/step.py | jq

🗋  tests/SampleJson.json
> The `SampleJson.json` has a pure JSON content in the payload field. This is 
> only the case if you have the Insight's parser step in front of your custom
> code, that parses the uploaded bytes upfront. If you have no parser step in
> front of your custom code the payload field contains `bytes` and no `dict`.

🗋  tests/SampleBinary.json
> If your step is the first in your pipeline, it will get the uploaded data in
> a structure like in this `SampleBinary.json`. The field payload contains the
> uploaded bytes as raw byte array, i.e. you get the original data without any
> manipulation or transforming through Iot Insights.

The `test.py` util has some more options. Find out more with 
```sh
$ python test.py -h
```  

## Deploy the code                                                                                 <a name="deploy"></a>

Once you have created a zip file, that contains the files `executable-manifest.yaml`, 
`src/insights_protocol.py` and `src/step.py`, you can upload it to your Bosch IoT Insights project as custom
 step under `Processing > Pipelines`.

The script `deploy_artifact.py` can help you to do the packaging and upload for you.
Open the file `deploy_artifact.py` and edit the variables `project`, `project_user`, `project_user_password` or use the
command line parameters of the script to overwrite the default values.
The username and password must be from a technical user with the **ADMIN role**.
Or you provide the credentials as environment variables, which is useful in a CI/CD environment.   
Environment variables:  

* INSIGHTS_PROJECT
* INSIGHTS_PROJECT_USER
* INSIGHTS_PROJECT_USER_PASSWORD

Then you can run the script with the command:
```sh
$ python deploy_artifact.py
```  
It will automatically bump the version and set the modifiedAt metadata in your `executable-manifest.yaml` and create the zip 
file `processor_artifact.zip`. Then it uploads the artifact to your project.

Afterwards go to your pipeline and select the new version in the custom step.

To show the help of the script execute:
```sh
$ python deploy_artifact.py -h 
```  

## The exchange data format                                                                        <a name="data-format"></a>

The exchange data dictionary is provided as argument of the `invoke` method. You can modify some
parts of it (`documents` and `variables` part).

### Complete example
```json
{
  "_id": {
    "collection" : "example_input_data",
    "document"   : {"$oid": "c001d001cec00ffee0040a11"},
    "pipelineId" : "default_pipeline",
    "projectId"  : "example"
  },
  "documents": [
    {
      "_id": {"$oid": "c001d001cec00ffee0040a11"},
      "metaData": {
        "receivedAt": {"$date": "2020-02-18T17:53:32.768Z"}
      },
      "payload": {
        "temperature": {"v": 20.5, "u":"°C"},
        "ts": {"$date": "2020-02-18T17:53:31.953Z"}
      }
    }
  ],
  "outputRef": [],
  "trigger": "dataIngestor",
  "variables": {
    "original_format": "ASN.1"
  },
  "error": "Traceback (most recent call last):\n  File \"src\\step.py\", line 44, in process_stream\n    output = ((await self.invoke(input)) or input)\n  File \"src\\step.py\", line 68, in invoke\n    raise Exception('My Error')\nException: My Error\n"
}
```

#### Field `_id`
Uniquely identifies the current pipeline execution run and is read only.  
Contained fields:  
* `collection`: name of the collection of the trigger document 
* `document`: id of that trigger document
* `pipelineId`: refers to the currently executed pipeline of your project 
* `projectId`: refers to your project this pipeline is used in
  
Example:  
```json
{
  "collection" : "example_input_data",
  "document"   : {"$oid": "c001d001cec00ffee0040a11"},
  "pipelineId" : "default_pipeline",
  "projectId"  : "example"
}
```  

#### Field `documents`
A list of input documents. Typically only one document. If the previous step put multiple documents
in the list, you can process them all.  
This list can be modified to contain any documents, that should be forwarded to the next step.

Typical example:
```json
[
  {
    "_id": {"$oid": "c001d001cec00ffee0040a11"},
    "metaData": {
      "receivedAt": {"$date": "2020-02-18T17:53:32.768Z"}
    },
    "payload": {
      "temperature": {"v": 20.5, "u":"°C"},
      "ts": {"$date": "2020-02-18T17:53:31.953Z"}
    }
  }
]
```

In case the object storage is used and the input type in the manifest is set to 'insights-payloadRef+json'
the field 'payloadRef' with a object url can be included in the exchange data.
For this case the pipeline step need to handle the load of the data on his own.
Be beware that the payloadRef is only injected into the custom step if it is the first step on a pipeline.

Example with object link:
```json
[
  {
    "_id": {"$oid": "c001d001cec00ffee0040a11"},
    "metaData": {
      "receivedAt": {"$date": "2020-02-18T17:53:32.768Z"}
    },
    "payloadRef": "http://user:passwd@192.168.100.1:9000/project/input/2021-04-28/608975998f91de1725d5efc5"
  }
]
```

#### Field `outputRef`
List of documents this step has created or modified directly in the database.  
Only used, if a step itself accesses the database directly, e.g. updating documents 
via an own MongoClient. Without this field, Insights is not aware of the context of 
these changes.  
But if the changed documents are referenced in the outputRef field, Insights can detect 
the actual processing execution and trigger which caused these change. E.g. that allows 
to trace back from processing result documents back to original input document.  
Example:
```json
[
  {"collection": "example_processed_data", "id": {"$oid": "c001d001cec00ffee0000001"}}
]
```  

#### Field `trigger`
What has triggered the processing (read only). Can be one of:
* `dataIngestor`:  trigger was a freshly uploaded document
* `reprocessing`:  reprocessing of formerly uploaded document
* `scheduler`:     name reserved, not yet used 
* `test`:          can be used in tests

#### Field `variables`
Dictionary of variables, that can be used for communication with the other steps of the same
pipeline execution run.  
Example:  
The first step can say that the actual original format parsed was `ASN.1`:  
```json
{"original_format": "ASN.1"}
```  

#### Field `error`
Contains information about raised exceptions by the custom Python code (stacktrace or custom error message). 
Example of setting a custom error message on the exchange data in the Python code:
```python
return exchange_data_holder._replace(error = "custom error message")
```
 or
```python
raise Exception("custom error message")  
```
The resulting JSON to write to the pipeline then looks like
```json
"error": "Traceback (most recent call last):\n  File \"src\\step.py\", line 44, in process_stream\n    output = ((await self.invoke(input)) or input)\n  File \"src\\step.py\", line 68, in invoke\n    raise Exception('My Error')\nException: My Error\n"  
```  

### Extended JSON
The exchange data format is using the Extended JSON format to encode BSON in the background.
See which Python types can be used in the documents and variables:

* [Python BSON Documentation]( https://api.mongodb.com/python/current/api/bson/index.html )
* [Extended JSON Documentation]( https://docs.mongodb.com/manual/reference/mongodb-extended-json/ )

## Glossary                                                                                        <a name="glossary"></a>

### Trigger Document
The document, which - with its creation - triggered the current pipeline execution.  
To avoid triggering on all incoming documents, it can be defined what properties a document must
fulfill in the Insight's UI. Select Processing > Pipelines > Pipeline > Configuration > Trigger
to configure a "Custom Filter".  
E.g.:  
Custom filter to trigger only on documents that haven't been uploaded before:
```json
{"tags": {"$not": {"$elemMatch": {"$eq": "duplicate"}}}}
```  

### Pipeline  
A trigger point and a chain of steps that work on the data of your project.  
Typically, the trigger is an newly uploaded document, which starts as just a byte blob, which
each step more and more transforms, enriches and harmonizes, until a meaningful document can
be stored by the output step into the processed data section of the database.  
