#!/usr/bin/python3

"""Demonstrates the use of MongoDB upserts and ExchangeDataHolder.outputRef

Uses data model and failure handling of insights_protocol.
Writes to MongoDB via async driver to not block the asyncio event queue.
https://www.mongodb.com/blog/post/getting-started-with-python-and-mongodb

"""

import sys
import os

from insights_protocol import StepBase, ExchangeDataHolder, MongoDBRef

import motor.motor_asyncio
from pymongo import InsertOne, DeleteOne, ReplaceOne, UpdateOne


mongo_uri = os.environ['MONGO_URI']
# MONGO_URI - this environment variable is made available in Insights' processing pipeline runtime environment.
#             I.e. it's already there and points to the Insights' project database.
#             For local testing, set up and use a local MongoDb. Using docker, two simple commands:
#                 docker create --name mongo -p 27017:27017 mongo
#                 docker start mongo
#             The required MONGO_URI value is then
#                 mongodb://localhost:27017/sfde?directConnection=true
#             If the local MongoDb already exists and if it is configured in replicaset mode, the value
#             needs to include the replicaset name like this:
#                 mongodb://localhost:27017/sfde?replicaSet=rs0&directConnection=true

client = motor.motor_asyncio.AsyncIOMotorClient(mongo_uri)
db = client.get_default_database()

class ExampleMongoDbUpserts(StepBase):

    async def invoke(self, exchange_data_holder: ExchangeDataHolder) -> ExchangeDataHolder:
        ids = [ doc['_id'] for doc in exchange_data_holder.documents if doc.get('_id') ]
        updates = [ UpdateOne({'_id':id}, {'$inc':{'c':1}}, upsert=True) for id in ids ]
        await db.example_output.bulk_write(updates)
        return exchange_data_holder._replace(
                                                documents=[],
                                                outputRef=[MongoDBRef(id, 'example_output') for id in ids],
                                            )

if __name__ == "__main__":
    input = sys.stdin.buffer
    output = sys.stdout.buffer
    sys.stdout = sys.stderr  # make any print in code going to stderr
    ExampleMongoDbUpserts(input,output).process_stream()
