#!/usr/bin/python3
from urllib.parse import urlparse, ParseResult, unquote

import boto3
from botocore.config import Config
import sys
from insights_protocol import StepBase, ExchangeDataHolder


# To make usage of the object storage you need to change the manifest entry 'input' from:
# insights+json
#     to
# insights-payloadRef+json
#
# Notes:
# Please beware that if you make usage of the object storage the custom step which handle the payloadRef
# should be on the first place of the pipeline.
# If the entry 'input' is missing on the manifest this will handheld like 'insights+json' is set.
#
# This Example:
# On this example we have create a jpeg (test/SampleJpeg.jpeg) where a comment
# with the text 'Insights rocks' exists.
# ----------------------------------------------------------------------------------
# | 00000000  ff d8 ff e0 00 10 4a 46  49 46 00 01 01 02 00 58  |ÿØÿà..JFIF.....X| |
# |                                     ...                                        |
# | 000031f0  01 01 01 01 01 01 01 01  01 ff fe 00 11 49 6e 73  |.........ÿþ..Ins| |
# | 00003200  69 67 68 74 73 20 72 6f  63 6b 73 00 ff c2 00 11  |ights rocks.ÿÂ..| |
# |                                     ...                                        |
# ----------------------------------------------------------------------------------
# We extract only the comment without loading the whole file and return just this text.
class ReadObjectFromS3(StepBase):
    # Called for every input
    async def invoke(self, exchange_data_holder: ExchangeDataHolder) -> ExchangeDataHolder:
        print("\nStart processing " + str(exchange_data_holder))
        first_doc = exchange_data_holder.documents[0]

        # If the input is set to insights-payloadRef+json the pipeline need to handle both types of inputs
        # - payloadRef contains a object store link for big objects
        # - payload containes the data for small objects
        if (first_doc.get('payload') is None):  # Case for big objects where a obect link is injected into the pipeline
            if (first_doc.get('payloadRef') is None):
                raise Exception('No data and no object store link is injected')

            # Example url http://user:passwd@192.168.100.1:9000/project/input/2021-04-28/608975998f91de1725d5efc5
            url: ParseResult = urlparse(first_doc['payloadRef'])
            s3_shim: S3Adapter = S3Adapter.instance(url)
            # For big objects not read the whole in one step
            # This can lead to a out of memory exception
            # If the object is to big try to process the object in smaller chunks to use less memory
            # Note: If you have still have memory problems be beware how python handle the memory
            # 1. Memory that was used from python and released again is rare returned to the underlay system
            #   -> If you have problems, try not to use too much memory at once, even if you release it
            # 2. As last try you can manual start the gc but have point one in mind
            # https://rushter.com/blog/python-memory-managment/
            jpegCommentStart: int = 0x31fd # Fix position of the comment in SampleJpeg.jpeg
            comment: bytes = s3_shim.readChunkOfObject(url, jpegCommentStart, jpegCommentStart + 14)
            comment: str = self.decodeBytesToString(comment)

            first_doc['payload'] = comment
            del first_doc['payloadRef']
        else:  # The case for small data which injected direct into the pipeline
            payload: bytes = first_doc['payload']
            comment: bytes = payload[0x31fd:(0x31fd + 14)]
            comment: str = self.decodeBytesToString(comment)
            first_doc['payload'] = comment

        print("processed %s\n" % exchange_data_holder.id)
        return exchange_data_holder

    # Modify the data
    def decodeBytesToString(self, object) -> str:
        # Do some operations on the object like here decode the bytes
        return object.decode('UTF-8');

class S3Adapter(object):
    __instance = None

    @staticmethod
    def instance(url: ParseResult):
        if S3Adapter.__instance == None:
            S3Adapter(url)
        elif not S3Adapter.__instance.isClientForSameUrl(url):
            raise Exception("URL '" + url + "' points to another instance than expected")
        return S3Adapter.__instance

    def __init__(self, url: ParseResult):
        if S3Adapter.__instance != None:
            raise Exception("Use S3Shim.instance(url) for reuse of s3 client")
        self.url: ParseResult = url
        self.bucket: str = self.__parseBucketFromUrl(url)
        self.s3_client: boto3.client = self.__createReadOnlyS3ClientFromUrl()
        S3Adapter.__instance = self

    def __parseBucketFromUrl(self, url: ParseResult) -> str:
        return url.path.split('/')[1]

    def __parseKeyFromUrl(self, url: ParseResult) -> str:
        return url.path.replace('/' + self.bucket + '/', '')

    def __createReadOnlyS3ClientFromUrl(self) -> boto3.client:
        # The password is url encoded so we need to decode it
        # before we can use it with the s3 client
        session = boto3.Session(
            aws_access_key_id=unquote(self.url.username),
            aws_secret_access_key=unquote(self.url.password))
        config = Config( retries = {
            'max_attempts': 3,
            'mode': 'standard'})
        apiUrl = '{}://{}:{}'.format(self.url.scheme,
                                     self.url.hostname,
                                     self.url.port)
        return session.client('s3', endpoint_url=apiUrl,config=config)

    def isClientForSameUrl(self, url: ParseResult) -> bool:
        bucket: str = self.__parseBucketFromUrl(url)
        return (url.hostname == self.url.hostname and
                url.username == self.url.username and
                url.password == self.url.password and
                bucket == self.bucket)

    # Use this method only for small files which fit into memory
    def readCompleteObject(self, url: ParseResult) -> bytes:
        resp = self.s3_client.get_object(Bucket=self.bucket,
                                         Key=self.__parseKeyFromUrl(url))
        return resp['Body'].read()

    def readChunkOfObject(self, url: ParseResult, start, end) -> bytes:
        resp = self.s3_client.get_object(Bucket=self.bucket,
                                         Key=self.__parseKeyFromUrl(url),
                                         Range='bytes={}-{}'.format(start, end - 1))
        return resp['Body'].read()

    def getSizeOfObjectInBytes(self, url: ParseResult) -> int:
        resp = self.s3_client.head_object(Bucket=self.bucket,
                                          Key=self.__parseKeyFromUrl(url))
        return resp['ContentLength']

if __name__ == "__main__":
    input = sys.stdin.buffer
    output = sys.stdout.buffer
    sys.stdout = sys.stderr  # make any print in code going to stderr
    ReadObjectFromS3(input, output).process_stream()
